"""Utilities to bulk load data into pgstac from json/ndjson."""
import contextlib
import itertools
import logging
import sys
import time
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Dict,
    Generator,
    Iterable,
    Iterator,
    Optional,
    TextIO,
    Tuple,
    Union,
)

import orjson
import psycopg
from cachetools.func import lru_cache
from orjson import JSONDecodeError
from plpygis.geometry import Geometry
from psycopg import sql
from smart_open import open
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)
from version_parser import Version as V

from .db import PgstacDB
from .hydration import dehydrate
from .version import __version__

logger = logging.getLogger(__name__)


@dataclass
class Partition:
    name: str
    collection: str
    datetime_range_min: str
    datetime_range_max: str
    end_datetime_range_min: str
    end_datetime_range_max: str
    requires_update: bool


def chunked_iterable(iterable: Iterable, size: Optional[int] = 10000) -> Iterable:
    """Chunk an iterable."""
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, size))
        if not chunk:
            break
        yield chunk


class Tables(str, Enum):
    """Available tables for loading."""

    items = "items"
    collections = "collections"


class Methods(str, Enum):
    """Available methods for loading data."""

    insert = "insert"
    ignore = "ignore"
    upsert = "upsert"
    delsert = "delsert"
    insert_ignore = "insert_ignore"


@contextlib.contextmanager
def open_std(
    filename: str, mode: str = "r", *args: Any, **kwargs: Any,
) -> Generator[Any, None, None]:
    """Open files and i/o streams transparently."""
    fh: Union[TextIO, BinaryIO]
    if (
        filename is None
        or filename == "-"
        or filename == "stdin"
        or filename == "stdout"
    ):
        stream = sys.stdin if "r" in mode else sys.stdout
        fh = stream.buffer if "b" in mode else stream
        close = False
    else:
        fh = open(filename, mode, *args, **kwargs)
        close = True

    try:
        yield fh
    finally:
        if close:
            try:
                fh.close()
            except AttributeError:
                pass


def read_json(file: Union[Path, str, Iterator[Any]] = "stdin") -> Iterable:
    """Load data from an ndjson or json file."""
    if file is None:
        file = "stdin"
    if isinstance(file, str):
        open_file: Any = open_std(file, "r")
        with open_file as f:
            # Try reading line by line as ndjson
            try:
                for line in f:
                    lineout = line.strip().replace("\\\\", "\\").replace("\\\\", "\\")
                    yield orjson.loads(lineout)
            except JSONDecodeError:
                # If reading first line as json fails, try reading entire file
                logger.info("First line could not be parsed as json, trying full file.")
                try:
                    f.seek(0)
                    json = orjson.loads(f.read())
                    if isinstance(json, list):
                        for record in json:
                            yield record
                    else:
                        yield json
                except JSONDecodeError:
                    logger.info("File cannot be read as json")
                    raise
    elif isinstance(file, Iterable):
        for line in file:
            if isinstance(line, Dict):
                yield line
            else:
                yield orjson.loads(line)


class Loader:
    """Utilities for loading data."""

    db: PgstacDB
    _partition_cache: Dict[str, Partition]

    def __init__(self, db: PgstacDB):
        self.db = db
        self._partition_cache: Dict[str, Partition] = {}

    def check_version(self) -> None:
        db_version = self.db.version
        if db_version is None:
            raise Exception("Failed to detect the target database version.")

        if db_version != "unreleased":
            v1 = V(db_version)
            v2 = V(__version__)
            if (v1.get_major_version(), v1.get_minor_version()) != (
                v2.get_major_version(),
                v2.get_minor_version(),
            ):
                raise Exception(
                    f"pypgstac version {__version__}"
                    " is not compatible with the target"
                    f" database version {self.db.version}."
                    f" database version {db_version}.",
                )

    @lru_cache(maxsize=128)
    def collection_json(self, collection_id: str) -> Tuple[Dict[str, Any], int, str]:
        """Get collection."""
        res = self.db.query_one(
            "SELECT base_item, key, partition_trunc FROM collections WHERE id=%s",
            (collection_id,),
        )
        if isinstance(res, tuple):
            base_item, key, partition_trunc = res
        else:
            raise Exception(f"Error getting info for {collection_id}.")
        if key is None:
            raise Exception(
                f"Collection {collection_id} is not present in the database",
            )
        logger.debug(f"Found {collection_id} with base_item {base_item}")
        return base_item, key, partition_trunc

    def load_collections(
        self,
        file: Union[Path, str, Iterator[Any]] = "stdin",
        insert_mode: Optional[Methods] = Methods.insert,
    ) -> None:
        """Load a collections json or ndjson file."""
        self.check_version()

        if file is None:
            file = "stdin"
        conn = self.db.connect()
        with conn.cursor() as cur:
            with conn.transaction():
                cur.execute(
                    """
                    DROP TABLE IF EXISTS tmp_collections;
                    CREATE TEMP TABLE tmp_collections
                    (content jsonb) ON COMMIT DROP;
                    """,
                )
                with cur.copy("COPY tmp_collections (content) FROM stdin;") as copy:
                    for collection in read_json(file):
                        copy.write_row((orjson.dumps(collection).decode(),))
                if insert_mode in (
                    None,
                    Methods.insert,
                ):
                    cur.execute(
                        """
                        INSERT INTO collections (content)
                        SELECT content FROM tmp_collections;
                        """,
                    )
                    logger.debug(cur.statusmessage)
                    logger.debug(f"Rows affected: {cur.rowcount}")
                elif insert_mode in (
                    Methods.insert_ignore,
                    Methods.ignore,
                ):
                    cur.execute(
                        """
                        INSERT INTO collections (content)
                        SELECT content FROM tmp_collections
                        ON CONFLICT DO NOTHING;
                        """,
                    )
                    logger.debug(cur.statusmessage)
                    logger.debug(f"Rows affected: {cur.rowcount}")
                elif insert_mode == Methods.upsert:
                    cur.execute(
                        """
                        INSERT INTO collections (content)
                        SELECT content FROM tmp_collections
                        ON CONFLICT (id) DO
                        UPDATE SET content=EXCLUDED.content;
                        """,
                    )
                    logger.debug(cur.statusmessage)
                    logger.debug(f"Rows affected: {cur.rowcount}")
                else:
                    raise Exception(
                        "Available modes are insert, ignore, and upsert."
                        f"You entered {insert_mode}.",
                    )

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_random_exponential(multiplier=1, max=120),
        retry=(
            retry_if_exception_type(psycopg.errors.CheckViolation)
            | retry_if_exception_type(psycopg.errors.DeadlockDetected)
        ),
        reraise=True,
    )
    def load_partition(
        self,
        partition: Partition,
        items: Iterable[Dict[str, Any]],
        insert_mode: Optional[Methods] = Methods.insert,
    ) -> None:
        """Load items data for a single partition."""
        conn = self.db.connect()
        t = time.perf_counter()

        logger.debug(f"Loading data for partition: {partition}.")
        with conn.cursor() as cur:
            if partition.requires_update:
                with conn.transaction():
                    cur.execute(
                        """
                        SELECT check_partition(
                            %s,
                            tstzrange(%s, %s, '[]'),
                            tstzrange(%s, %s, '[]')
                        );
                    """,
                        (
                            partition.collection,
                            partition.datetime_range_min,
                            partition.datetime_range_max,
                            partition.end_datetime_range_min,
                            partition.end_datetime_range_max,
                        ),
                    )

                    logger.debug(
                        f"Adding or updating partition {partition.name} "
                        f"took {time.perf_counter() - t}s",
                    )
                partition.requires_update = False
            else:
                logger.debug(f"Partition {partition.name} does not require an update.")

            with conn.transaction():

                t = time.perf_counter()
                if insert_mode in (
                    None,
                    Methods.insert,
                ):
                    with cur.copy(
                        sql.SQL(
                            """
                            COPY {}
                            (id, collection, datetime, end_datetime, geometry, content)
                            FROM stdin;
                            """,
                        ).format(sql.Identifier(partition.name)),
                    ) as copy:
                        for item in items:
                            item.pop("partition")
                            copy.write_row(
                                (
                                    item["id"],
                                    item["collection"],
                                    item["datetime"],
                                    item["end_datetime"],
                                    item["geometry"],
                                    item["content"],
                                ),
                            )
                    logger.debug(cur.statusmessage)
                    logger.debug(f"Rows affected: {cur.rowcount}")
                elif insert_mode in (
                    Methods.insert_ignore,
                    Methods.upsert,
                    Methods.delsert,
                    Methods.ignore,
                ):
                    cur.execute(
                        """
                        DROP TABLE IF EXISTS items_ingest_temp;
                        CREATE TEMP TABLE items_ingest_temp
                        ON COMMIT DROP AS SELECT * FROM items LIMIT 0;
                        """,
                    )
                    with cur.copy(
                        """
                        COPY items_ingest_temp
                        (id, collection, datetime, end_datetime, geometry, content)
                        FROM stdin;
                        """,
                    ) as copy:
                        for item in items:
                            item.pop("partition")
                            copy.write_row(
                                (
                                    item["id"],
                                    item["collection"],
                                    item["datetime"],
                                    item["end_datetime"],
                                    item["geometry"],
                                    item["content"],
                                ),
                            )
                    logger.debug(cur.statusmessage)
                    logger.debug(f"Copied rows: {cur.rowcount}")

                    cur.execute(
                        sql.SQL(
                            """
                                LOCK TABLE ONLY {} IN EXCLUSIVE MODE;
                            """,
                        ).format(sql.Identifier(partition.name)),
                    )
                    if insert_mode in (
                        Methods.ignore,
                        Methods.insert_ignore,
                    ):
                        cur.execute(
                            sql.SQL(
                                """
                                INSERT INTO {}
                                SELECT *
                                FROM items_ingest_temp ON CONFLICT DO NOTHING;
                                """,
                            ).format(sql.Identifier(partition.name)),
                        )
                        logger.debug(cur.statusmessage)
                        logger.debug(f"Rows affected: {cur.rowcount}")
                    elif insert_mode == Methods.upsert:
                        cur.execute(
                            sql.SQL(
                                """
                                INSERT INTO {} AS t SELECT * FROM items_ingest_temp
                                ON CONFLICT (id) DO UPDATE
                                SET
                                    datetime = EXCLUDED.datetime,
                                    end_datetime = EXCLUDED.end_datetime,
                                    geometry = EXCLUDED.geometry,
                                    collection = EXCLUDED.collection,
                                    content = EXCLUDED.content
                                WHERE t IS DISTINCT FROM EXCLUDED
                                ;
                            """,
                            ).format(sql.Identifier(partition.name)),
                        )
                        logger.debug(cur.statusmessage)
                        logger.debug(f"Rows affected: {cur.rowcount}")
                    elif insert_mode == Methods.delsert:
                        cur.execute(
                            sql.SQL(
                                """
                                WITH deletes AS (
                                    DELETE FROM items i USING items_ingest_temp s
                                        WHERE
                                            i.id = s.id
                                            AND i.collection = s.collection
                                )
                                INSERT INTO {} AS t SELECT * FROM items_ingest_temp
                                ON CONFLICT (id) DO UPDATE
                                SET
                                    datetime = EXCLUDED.datetime,
                                    end_datetime = EXCLUDED.end_datetime,
                                    geometry = EXCLUDED.geometry,
                                    collection = EXCLUDED.collection,
                                    content = EXCLUDED.content
                                WHERE t IS DISTINCT FROM EXCLUDED
                                ;
                                """,
                            ).format(sql.Identifier(partition.name)),

                        )
                        logger.debug(cur.statusmessage)
                        logger.debug(f"Rows affected: {cur.rowcount}")
                else:
                    raise Exception(
                        "Available modes are insert, ignore, upsert, and delsert."
                        f"You entered {insert_mode}.",
                    )
                logger.debug("Updating Partition Stats")
                cur.execute("SELECT update_partition_stats_q(%s);",(partition.name,))
                logger.debug(cur.statusmessage)
                logger.debug(f"Rows affected: {cur.rowcount}")
        logger.debug(
            f"Copying data for {partition} took {time.perf_counter() - t} seconds",
        )

    def _partition_update(self, item: Dict[str, Any]) -> str:
        """Update the cached partition with the item information and return the name.

        This method will mark the partition as dirty if the bounds of the partition
        need to be updated based on this item.
        """
        p = item.get("partition", None)
        if p is None:
            _, key, partition_trunc = self.collection_json(item["collection"])
            if partition_trunc == "year":
                pd = item["datetime"].replace("-", "")[:4]
                p = f"_items_{key}_{pd}"
            elif partition_trunc == "month":
                pd = item["datetime"].replace("-", "")[:6]
                p = f"_items_{key}_{pd}"
            else:
                p = f"_items_{key}"
            item["partition"] = p

        partition_name: str = p

        partition: Optional[Partition] = None

        if partition_name not in self._partition_cache:
            # Read the partition information from the database if it exists
            db_rows = list(
                self.db.query(
                    """
                    SELECT
                        nullif(lower(constraint_dtrange),'-infinity')
                            as datetime_range_min,
                        nullif(upper(constraint_dtrange),'infinity')
                            as datetime_range_max,
                        nullif(lower(constraint_edtrange),'-infinity')
                            as end_datetime_range_min,
                        nullif(upper(constraint_edtrange),'infinity')
                            as end_datetime_range_max
                    FROM partitions WHERE partition=%s;
                    """,
                    [partition_name],
                ),
            )
            if db_rows:
                datetime_range_min: Optional[datetime] = db_rows[0][0] or datetime.min
                datetime_range_max: Optional[datetime] = db_rows[0][1] or datetime.max
                end_datetime_range_min: Optional[datetime] = (
                    db_rows[0][2] or datetime.min
                )
                end_datetime_range_max: Optional[datetime] = (
                    db_rows[0][3] or datetime.max
                )

                partition = Partition(
                    name=partition_name,
                    collection=item["collection"],
                    datetime_range_min=datetime_range_min.isoformat(),
                    datetime_range_max=datetime_range_max.isoformat(),
                    end_datetime_range_min=end_datetime_range_min.isoformat(),
                    end_datetime_range_max=end_datetime_range_max.isoformat(),
                    requires_update=False,
                )

        else:
            partition = self._partition_cache[partition_name]

        if partition:
            # Only update the partition if the item is outside the current bounds
            if item["datetime"] < partition.datetime_range_min:
                partition.datetime_range_min = item["datetime"]
                partition.requires_update = True
            if item["datetime"] > partition.datetime_range_max:
                partition.datetime_range_max = item["datetime"]
                partition.requires_update = True
            if item["end_datetime"] < partition.end_datetime_range_min:
                partition.end_datetime_range_min = item["end_datetime"]
                partition.requires_update = True
            if item["end_datetime"] > partition.end_datetime_range_max:
                partition.end_datetime_range_max = item["end_datetime"]
                partition.requires_update = True
        else:
            # No partition exists yet; create a new one from item
            partition = Partition(
                name=partition_name,
                collection=item["collection"],
                datetime_range_min=item["datetime"],
                datetime_range_max=item["datetime"],
                end_datetime_range_min=item["end_datetime"],
                end_datetime_range_max=item["end_datetime"],
                requires_update=True,
            )

        self._partition_cache[partition_name] = partition

        return partition_name

    def read_dehydrated(self, file: Union[Path, str] = "stdin") -> Generator:
        if file is None:
            file = "stdin"
        if isinstance(file, str):
            open_file: Any = open_std(file, "r")
            with open_file as f:
                # Note: if 'content' is changed to be anything
                # but the last field, the logic below will break.
                fields = [
                    "id",
                    "geometry",
                    "collection",
                    "datetime",
                    "end_datetime",
                    "content",
                ]

                for line in f:
                    tab_split = line.split("\t")
                    item = {}
                    for i, field in enumerate(fields):
                        if field == "content":
                            # Join the remaining splits in case
                            # there were any tabs in the JSON content.
                            content_value = "\t".join(tab_split[i:])
                            # Replace quote characters that can be
                            # written on export and causes failures.
                            content_value = content_value.replace(r'\\"', r"\"")
                            item[field] = content_value
                        else:
                            item[field] = tab_split[i]
                    item["partition"] = self._partition_update(item)
                    yield item

    def read_hydrated(
        self, file: Union[Path, str, Iterator[Any]] = "stdin",
    ) -> Generator:
        for line in read_json(file):
            item = self.format_item(line)
            item["partition"] = self._partition_update(item)
            yield item

    def load_items(
        self,
        file: Union[Path, str, Iterator[Any]] = "stdin",
        insert_mode: Optional[Methods] = Methods.insert,
        dehydrated: Optional[bool] = False,
        chunksize: Optional[int] = 10000,
    ) -> None:
        """Load items json records."""
        self.check_version()

        if file is None:
            file = "stdin"
        t = time.perf_counter()
        self._partition_cache = {}

        if dehydrated and isinstance(file, str):
            items = self.read_dehydrated(file)
        else:
            items = self.read_hydrated(file)

        for chunkin in chunked_iterable(items, chunksize):
            chunk = list(chunkin)
            chunk.sort(key=lambda x: x["partition"])
            for k, g in itertools.groupby(chunk, lambda x: x["partition"]):
                self.load_partition(self._partition_cache[k], g, insert_mode)

        logger.debug(f"Adding data to database took {time.perf_counter() - t} seconds.")

    def format_item(self, _item: Union[Path, str, Dict[str, Any]]) -> Dict[str, Any]:
        """Format an item to insert into a record."""
        out: Dict[str, Any] = {}
        item: Dict[str, Any]
        if not isinstance(_item, dict):
            try:
                item = orjson.loads(str(_item).replace("\\\\", "\\"))
            except Exception:
                raise
        else:
            item = _item

        base_item, key, partition_trunc = self.collection_json(item["collection"])

        out["id"] = item.get("id")
        out["collection"] = item.get("collection")
        properties: Dict[str, Any] = item.get("properties", {})

        dt: Optional[str] = properties.get("datetime")
        edt: Optional[str] = properties.get("end_datetime")
        sdt: Optional[str] = properties.get("start_datetime")

        if edt is not None and sdt is not None:
            out["datetime"] = sdt
            out["end_datetime"] = edt
        elif dt is not None:
            out["datetime"] = dt
            out["end_datetime"] = dt
        else:
            raise Exception("Invalid datetime encountered")

        if out["datetime"] is None or out["end_datetime"] is None:
            raise Exception(
                f"Datetime must be set. OUT: {out} Properties: {properties}",
            )

        if partition_trunc == "year":
            pd = out["datetime"].replace("-", "")[:4]
            partition = f"_items_{key}_{pd}"
        elif partition_trunc == "month":
            pd = out["datetime"].replace("-", "")[:6]
            partition = f"_items_{key}_{pd}"
        else:
            partition = f"_items_{key}"

        out["partition"] = partition

        geojson = item.get("geometry")
        if geojson is None:
            geometry = None
        else:
            geom = Geometry.from_geojson(geojson)
            if geom is None:
                raise Exception(f"Invalid geometry encountered: {geojson}")
            geometry = str(geom.wkb)
        out["geometry"] = geometry

        content = dehydrate(base_item, item)

        # Remove keys from the dehydrated item content which are stored directly
        # on the table row.
        content.pop("id", None)
        content.pop("collection", None)
        content.pop("geometry", None)

        out["content"] = orjson.dumps(content).decode()

        return out

    def __hash__(self) -> int:
        """Return hash so that the LRU deocrator can cache without the class."""
        return 0



"""Fixtures for pypgstac tests."""
import os
from typing import Generator

import psycopg
import pytest

from pypgstac.db import PgstacDB
from pypgstac.load import Loader
from pypgstac.migrate import Migrate


@pytest.fixture(scope="function")
def db() -> Generator:
    """Fixture to get a fresh database."""
    origdb: str = os.getenv("PGDATABASE", "")

    with psycopg.connect(autocommit=True) as conn:
        try:
            conn.execute(
                """
                CREATE DATABASE pypgstactestdb
                TEMPLATE pgstac_test_db_template;
                """,
            )
        except psycopg.errors.DuplicateDatabase:
            try:
                conn.execute(
                    """
                    DROP DATABASE pypgstactestdb WITH (FORCE);
                    """,
                )
                conn.execute(
                    """
                    CREATE DATABASE pypgstactestdb
                    TEMPLATE pgstac_test_db_template;
                    """,
                )
            except psycopg.errors.InsufficientPrivilege:
                try:
                    conn.execute("DROP DATABASE pypgstactestdb;")
                    conn.execute(
                    """
                    CREATE DATABASE pypgstactestdb
                    TEMPLATE pgstac_test_db_template;
                    """,
                )
                except Exception:
                    pass

    os.environ["PGDATABASE"] = "pypgstactestdb"

    pgdb = PgstacDB()

    yield pgdb

    pgdb.close()
    os.environ["PGDATABASE"] = origdb

    with psycopg.connect(autocommit=True) as conn:
        try:
            conn.execute("DROP DATABASE pypgstactestdb WITH (FORCE);")
        except psycopg.errors.InsufficientPrivilege:
            try:
                conn.execute("DROP DATABASE pypgstactestdb;")
            except Exception:
                pass


@pytest.fixture(scope="function")
def loader(db: PgstacDB) -> Generator:
    """Fixture to get a loader and an empty pgstac."""
    if False:
        db.query("DROP SCHEMA IF EXISTS pgstac CASCADE;")
        Migrate(db).run_migration()
    ldr = Loader(db)
    return ldr



"""Hydrate data in pypgstac rather than on the database."""
from copy import deepcopy
from typing import Any, Dict

from pypgstac.pgstacrs import hydrate  # noqa # pylint: disable=unused-import

# Marker value to indicate that a key should not be rehydrated
DO_NOT_MERGE_MARKER = "𒍟※"

def hydrate_py(base_item: Dict[str, Any], item: Dict[str, Any]) -> Dict[str, Any]:
    """Hydrate item in-place with base_item properties.

    This will not perform a deep copy; values of the original item will be referenced
    in the return item.
    """
    # Merge will mutate i, but create deep copies of values in the base item
    # This will prevent the base item values from being mutated, e.g. by
    # filtering out fields in `filter_fields`.
    def merge(b: Dict[str, Any], i: Dict[str, Any]) -> None:
        for key in b:
            if key in i:
                if isinstance(b[key], dict) and isinstance(i.get(key), dict):
                    # Recurse on dicts to merge values
                    merge(b[key], i[key])
                elif isinstance(b[key], list) and isinstance(i.get(key), list):
                    # Merge unequal lists, assume uniform types
                    if len(b[key]) == len(i[key]):
                        for bb, ii in zip(b[key], i[key]):
                            # Make sure we're merging two dicts
                            if isinstance(bb, dict) and isinstance(ii, dict):
                                merge(bb, ii)
                    else:
                        # If item has a different length, then just use the item value
                        continue
                else:
                    # Key exists on item but isn't a dict or list, keep item value
                    if i[key] == DO_NOT_MERGE_MARKER:
                        # Key was marked as do-not-merge, drop it from the item
                        del i[key]
                    else:
                        # Keep the item value
                        continue

            else:
                # Keys in base item that are not in item are simply copied over
                i[key] = deepcopy(b[key])

    merge(base_item, item)
    return item


def dehydrate(base_item: Dict[str, Any], full_item: Dict[str, Any]) -> Dict[str, Any]:
    """
    Get a recursive difference between a base item and an incoming item to dehydrate.

    For keys of dicts within items, if the base item contains a key not present
    in the incoming item, then a do-no-merge value is added indicating that the
    key should not be rehydrated with the corresponding base item value. This will allow
    collection item-assets to contain keys that may not be present on individual items.
    """

    def strip(base_value: Dict[str, Any], item_value: Dict[str, Any]) -> Dict[str, Any]:
        out: dict = {}
        for key, value in item_value.items():
            if base_value is None or key not in base_value:
                # Nothing on base; preserve item value in the dehydrated item
                out[key] = value
                continue

            if base_value[key] == value:
                # Equal values; do not include in the dehydrated item
                continue

            if isinstance(base_value[key], list) and isinstance(value, list):
                if len(base_value[key]) == len(value):
                    # Equal length lists dehydrate dicts at each matching index
                    # and use incoming item values for other types
                    out[key] = []
                    for bv, v in zip(base_value[key], value):
                        if isinstance(bv, dict) and isinstance(v, dict):
                            dehydrated = strip(bv, v)
                            apply_marked_keys(bv, v, dehydrated)
                            out[key].append(dehydrated)
                        else:
                            out[key].append(v)
                else:
                    # Unequal length lists are not dehydrated and just use the
                    # incoming item value
                    out[key] = value
                continue

            if value is None or value == []:
                # Don't keep empty values
                continue

            if isinstance(value, dict):
                # After dehdrating a dict, mark any keys that are present on the
                # base item but not in the incoming item as `do-not-merge` during
                # rehydration
                dehydrated = strip(base_value[key], value)
                apply_marked_keys(base_value[key], value, dehydrated)
                out[key] = dehydrated
                continue
            else:
                # Unequal non-dict values are copied over from the incoming item
                out[key] = value

        # Mark any top-level keys from the base_item that are not in the incoming item
        apply_marked_keys(base_value, item_value, out)

        return out

    return strip(base_item, full_item)


def apply_marked_keys(
    base_item: Dict[str, Any],
    full_item: Dict[str, Any],
    dehydrated: Dict[str, Any],
) -> None:
    """Mark keys.

    Mark any keys that are present on the base item but not in the incoming item
    as `do-not-merge` on the dehydrated item. This will prevent they key from
    being rehydrated.

    This modifies the dehydrated item in-place.
    """
    try:
        marked_keys = [key for key in base_item if key not in full_item.keys()]
        marked_dict = {k: DO_NOT_MERGE_MARKER for k in marked_keys}
        dehydrated.update(marked_dict)
    except TypeError:
        pass
